/*
 * MIT License
 *
 * Copyright (c) 2019 Alexey Edelev <semlanik@gmail.com>
 *
 * This file is part of NeuralNetwork project https://git.semlanik.org/semlanik/NeuralNetwork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this
 * software and associated documentation files (the "Software"), to deal in the Software
 * without restriction, including without limitation the rights to use, copy, modify,
 * merge, publish, distribute, sublicense, and/or sell copies of the Software, and
 * to permit persons to whom the Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies
 * or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

package mutagens

import (
	"math/rand"
	"time"

	neuralnetwork "../../neuralnetwork"
)

// DummyMutagen is simple randomized mutagen
type DummyMutagen struct {
	chance        float64
	mutationCount int
}

// NewDummyMutagen constructs DummyMutagen with specified mutation chance and
// amount of mutations that should be applied per cycle
func NewDummyMutagen(chance float64, mutationCount int) (dm *DummyMutagen) {
	dm = &DummyMutagen{
		chance:        chance,
		mutationCount: mutationCount,
	}
	return
}

// Mutate implementaion of Mutagen inteface Mutate method
// For DummyMutagen it gets pseudo-random number and validates if number in
// chance bounds. After method applies randomized mutation for random weight
// and bias in neuralnetwork.NeuralNetwork
func (dm *DummyMutagen) Mutate(network *neuralnetwork.NeuralNetwork) {
	rand.Seed(time.Now().UnixNano())
	for l := 1; l < network.LayerCount; l++ {
		randomized := rand.Float64()
		if randomized < dm.chance {
			r, c := network.Weights[l].Dims()
			for o := 0; o < dm.mutationCount; o++ {
				mutationRow := int(rand.Uint32()) % r
				mutationColumn := int(rand.Uint32()) % c
				weight := rand.NormFloat64()
				bias := rand.NormFloat64()
				network.Weights[l].Set(mutationRow, mutationColumn, weight)
				network.Biases[l].Set(mutationRow, 0, bias)
			}
		}
	}
}
